#pragma once

#include "gdconfig.hpp"
#include "weight.hpp"

namespace dichotomy {

/*!
 * \brief class to store the resul: solutions, weights and weight graph
 */
class Result {

    public:

        Result(std::vector<Weight>& w, const WeightGraph& g, const std::vector<dichotomy::CostVec>& s, size_t& n):
        weights(w), weightGraph(g), solutions(s), nSubSolve(n)
        {

        }

    public:

        const std::vector<dichotomy::CostVec>& solutions;
        const std::vector<Weight>& weights;
        const dichotomy::WeightGraph& weightGraph;
        size_t& nSubSolve; // nb of time sub problems are solved

};

}